@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library.  If not, see
- * <http://www.gnu.org/licenses/>.
+ * <https://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -53,7 +53,14 @@
 
 #ifndef G_OS_WIN32
 #include "libgimpbase/gimpsignal.h"
+
 #else
+
+#ifdef HAVE_EXCHNDL
+#include <time.h>
+#include <exchndl.h>
+#endif
+
 #include <signal.h>
 #endif
 
@@ -87,9 +94,18 @@
 #endif
 
 #if defined(G_OS_WIN32)
+#  ifdef STRICT
+#  undef STRICT
+#  endif
 #  define STRICT
+
+#  ifdef _WIN32_WINNT
+#  undef _WIN32_WINNT
+#  endif
 #  define _WIN32_WINNT 0x0601
+
 #  include <windows.h>
+#  include <tlhelp32.h>
 #  undef RGB
 #  define USE_WIN32_SHM 1
 #endif
@@ -101,7 +117,6 @@
 #include "libgimpbase/gimpprotocol.h"
 #include "libgimpbase/gimpwire.h"
 
-#undef GIMP_DISABLE_DEPRECATED
 #include "gimp.h"
 #include "gimpunitcache.h"
 
@@ -119,7 +134,7 @@
  **/
 
 
-#define TILE_MAP_SIZE (_tile_width * _tile_height * 4)
+#define TILE_MAP_SIZE (_tile_width * _tile_height * 32)
 
 #define ERRMSG_SHM_FAILED "Could not attach to gimp shared memory segment"
 
@@ -148,7 +163,15 @@ static void       gimp_message_func            (const gchar    *log_domain,
                                                 GLogLevelFlags  log_level,
                                                 const gchar    *message,
                                                 gpointer        data);
-#ifndef G_OS_WIN32
+static void       gimp_fatal_func              (const gchar    *log_domain,
+                                                GLogLevelFlags  flags,
+                                                const gchar    *message,
+                                                gpointer        data);
+#ifdef G_OS_WIN32
+#ifdef HAVE_EXCHNDL
+static LONG WINAPI gimp_plugin_sigfatal_handler (PEXCEPTION_POINTERS pExceptionInfo);
+#endif
+#else
 static void       gimp_plugin_sigfatal_handler (gint            sig_num);
 #endif
 static gboolean   gimp_plugin_io_error_handler (GIOChannel      *channel,
@@ -174,8 +197,13 @@ static void       gimp_set_pdb_error           (const GimpParam *return_vals,
                                                 gint             n_return_vals);
 
 
-static GIOChannel *_readchannel  = NULL;
-GIOChannel *_writechannel = NULL;
+#if defined G_OS_WIN32 && defined HAVE_EXCHNDL
+static LPTOP_LEVEL_EXCEPTION_FILTER  _prevExceptionFilter    = NULL;
+static gchar                         *plug_in_backtrace_path = NULL;
+#endif
+
+static GIOChannel                   *_readchannel            = NULL;
+GIOChannel                          *_writechannel           = NULL;
 
 #ifdef USE_WIN32_SHM
 static HANDLE shm_handle;
@@ -185,10 +213,12 @@ static gint           _tile_width        = -1;
 static gint           _tile_height       = -1;
 static gint           _shm_ID            = -1;
 static guchar        *_shm_addr          = NULL;
-static const gdouble  _gamma_val         = 2.2;
-static gboolean       _install_cmap      = FALSE;
 static gboolean       _show_tool_tips    = TRUE;
 static gboolean       _show_help_button  = TRUE;
+static gboolean       _export_profile    = FALSE;
+static gboolean       _export_exif       = FALSE;
+static gboolean       _export_xmp        = FALSE;
+static gboolean       _export_iptc       = FALSE;
 static GimpCheckSize  _check_size        = GIMP_CHECK_SIZE_MEDIUM_CHECKS;
 static GimpCheckType  _check_type        = GIMP_CHECK_TYPE_GRAY_CHECKS;
 static gint           _min_colors        = 144;
@@ -197,6 +227,7 @@ static gchar         *_wm_class          = NULL;
 static gchar         *_display_name      = NULL;
 static gint           _monitor_number    = 0;
 static guint32        _timestamp         = 0;
+static gchar         *_icon_theme_dir    = NULL;
 static const gchar   *progname           = NULL;
 
 static gchar          write_buffer[WRITE_BUFFER_SIZE];
@@ -244,6 +275,18 @@ gimp_main (const GimpPlugInInfo *info,
            gint                  argc,
            gchar                *argv[])
 {
+  enum
+  {
+    ARG_PROGNAME,
+    ARG_GIMP,
+    ARG_READ_FD,
+    ARG_WRITE_FD,
+    ARG_MODE,
+    ARG_STACK_TRACE_MODE,
+
+    N_ARGS
+  };
+
   gchar       *basename;
   const gchar *env_string;
   gchar       *debug_string;
@@ -256,8 +299,8 @@ gimp_main (const GimpPlugInInfo *info,
     typedef BOOL (WINAPI *t_SetDllDirectoryA) (LPCSTR lpPathName);
     t_SetDllDirectoryA p_SetDllDirectoryA;
 
-    p_SetDllDirectoryA = GetProcAddress (GetModuleHandle ("kernel32.dll"),
-					 "SetDllDirectoryA");
+    p_SetDllDirectoryA = (t_SetDllDirectoryA) GetProcAddress (GetModuleHandle ("kernel32.dll"),
+                                                              "SetDllDirectoryA");
     if (p_SetDllDirectoryA)
       (*p_SetDllDirectoryA) ("");
   }
@@ -295,18 +338,63 @@ gimp_main (const GimpPlugInInfo *info,
     g_free (bin_dir);
   }
 
+#ifdef HAVE_EXCHNDL
+  /* Use Dr. Mingw (dumps backtrace on crash) if it is available. */
+  {
+    time_t  t;
+    gchar  *filename;
+    gchar  *dir;
+
+    /* This has to be the non-roaming directory (i.e., the local
+       directory) as backtraces correspond to the binaries on this
+       system. */
+    dir = g_build_filename (g_get_user_data_dir (),
+                            GIMPDIR, GIMP_USER_VERSION, "CrashLog",
+                            NULL);
+    /* Ensure the path exists. */
+    g_mkdir_with_parents (dir, 0700);
+
+    time (&t);
+    filename = g_strdup_printf ("%s-crash-%" G_GUINT64_FORMAT ".txt",
+                                g_get_prgname(), t);
+    plug_in_backtrace_path = g_build_filename (dir, filename, NULL);
+    g_free (filename);
+    g_free (dir);
+
+    /* Similar to core crash handling in app/signals.c, the order here
+     * is very important!
+     */
+    if (! _prevExceptionFilter)
+      _prevExceptionFilter = SetUnhandledExceptionFilter (gimp_plugin_sigfatal_handler);
+
+    ExcHndlInit ();
+    ExcHndlSetLogFileNameA (plug_in_backtrace_path);
+  }
+#endif
+
 #ifndef _WIN64
   {
     typedef BOOL (WINAPI *t_SetProcessDEPPolicy) (DWORD dwFlags);
     t_SetProcessDEPPolicy p_SetProcessDEPPolicy;
 
     p_SetProcessDEPPolicy = GetProcAddress (GetModuleHandle ("kernel32.dll"),
-					    "SetProcessDEPPolicy");
+                                            "SetProcessDEPPolicy");
     if (p_SetProcessDEPPolicy)
       (*p_SetProcessDEPPolicy) (PROCESS_DEP_ENABLE|PROCESS_DEP_DISABLE_ATL_THUNK_EMULATION);
   }
 #endif
 
+  /* Group all our windows together on the taskbar */
+  {
+    typedef HRESULT (WINAPI *t_SetCurrentProcessExplicitAppUserModelID) (PCWSTR lpPathName);
+    t_SetCurrentProcessExplicitAppUserModelID p_SetCurrentProcessExplicitAppUserModelID;
+
+    p_SetCurrentProcessExplicitAppUserModelID = (t_SetCurrentProcessExplicitAppUserModelID) GetProcAddress (GetModuleHandle ("shell32.dll"),
+                                                                                                            "SetCurrentProcessExplicitAppUserModelID");
+    if (p_SetCurrentProcessExplicitAppUserModelID)
+      (*p_SetCurrentProcessExplicitAppUserModelID) (L"gimp.GimpApplication");
+  }
+
   /* Check for exe file name with spaces in the path having been split up
    * by buggy NT C runtime, or something. I don't know why this happens
    * on NT (including w2k), but not on w95/98.
@@ -325,7 +413,7 @@ gimp_main (const GimpPlugInInfo *info,
                */
               GString *s;
 
-              s = g_string_new (argv[0]);
+              s = g_string_new (argv[ARG_PROGNAME]);
 
               for (j = 1; j <= i; j++)
                 {
@@ -333,7 +421,7 @@ gimp_main (const GimpPlugInInfo *info,
                   s = g_string_append (s, argv[j]);
                 }
 
-              argv[0] = s->str;
+              argv[ARG_PROGNAME] = s->str;
 
               /* Move rest of argv down */
               for (j = 1; j < argc - i; j++)
@@ -352,16 +440,16 @@ gimp_main (const GimpPlugInInfo *info,
 
   PLUG_IN_INFO = *info;
 
-  if ((argc != 6) || (strcmp (argv[1], "-gimp") != 0))
+  if ((argc != N_ARGS) || (strcmp (argv[ARG_GIMP], "-gimp") != 0))
     {
       g_printerr ("%s is a GIMP plug-in and must be run by GIMP to be used\n",
-                  argv[0]);
+                  argv[ARG_PROGNAME]);
       return 1;
     }
 
   gimp_env_init (TRUE);
 
-  progname = argv[0];
+  progname = argv[ARG_PROGNAME];
 
   basename = g_path_get_basename (progname);
 
@@ -410,7 +498,7 @@ gimp_main (const GimpPlugInInfo *info,
 
   g_free (basename);
 
-  stack_trace_mode = (GimpStackTraceMode) CLAMP (atoi (argv[5]),
+  stack_trace_mode = (GimpStackTraceMode) CLAMP (atoi (argv[ARG_STACK_TRACE_MODE]),
                                                  GIMP_STACK_TRACE_NEVER,
                                                  GIMP_STACK_TRACE_ALWAYS);
 
@@ -423,9 +511,11 @@ gimp_main (const GimpPlugInInfo *info,
   gimp_signal_private (SIGHUP,  gimp_plugin_sigfatal_handler, 0);
   gimp_signal_private (SIGINT,  gimp_plugin_sigfatal_handler, 0);
   gimp_signal_private (SIGQUIT, gimp_plugin_sigfatal_handler, 0);
+  gimp_signal_private (SIGTERM, gimp_plugin_sigfatal_handler, 0);
+
+  gimp_signal_private (SIGABRT, gimp_plugin_sigfatal_handler, 0);
   gimp_signal_private (SIGBUS,  gimp_plugin_sigfatal_handler, 0);
   gimp_signal_private (SIGSEGV, gimp_plugin_sigfatal_handler, 0);
-  gimp_signal_private (SIGTERM, gimp_plugin_sigfatal_handler, 0);
   gimp_signal_private (SIGFPE,  gimp_plugin_sigfatal_handler, 0);
 
   /* Ignore SIGPIPE from crashing Gimp */
@@ -436,11 +526,11 @@ gimp_main (const GimpPlugInInfo *info,
 #endif
 
 #ifdef G_OS_WIN32
-  _readchannel  = g_io_channel_win32_new_fd (atoi (argv[2]));
-  _writechannel = g_io_channel_win32_new_fd (atoi (argv[3]));
+  _readchannel  = g_io_channel_win32_new_fd (atoi (argv[ARG_READ_FD]));
+  _writechannel = g_io_channel_win32_new_fd (atoi (argv[ARG_WRITE_FD]));
 #else
-  _readchannel  = g_io_channel_unix_new (atoi (argv[2]));
-  _writechannel = g_io_channel_unix_new (atoi (argv[3]));
+  _readchannel  = g_io_channel_unix_new (atoi (argv[ARG_READ_FD]));
+  _writechannel = g_io_channel_unix_new (atoi (argv[ARG_WRITE_FD]));
 #endif
 
   g_io_channel_set_encoding (_readchannel, NULL, NULL);
@@ -457,7 +547,6 @@ gimp_main (const GimpPlugInInfo *info,
   gimp_wire_set_writer (gimp_write);
   gimp_wire_set_flusher (gimp_flush);
 
-  g_type_init ();
   gimp_enums_init ();
 
   /*  initialize units  */
@@ -492,15 +581,31 @@ gimp_main (const GimpPlugInInfo *info,
 
 
   /* set handler both for the "LibGimp" and "" domains */
+  {
+    const gchar * const log_domains[] =
+    {
+      "LibGimp",
+      "LibGimpBase",
+      "LibGimpColor",
+      "LibGimpConfig",
+      "LibGimpMath",
+      "LibGimpModule",
+      "LibGimpThumb",
+      "LibGimpWidgets"
+    };
+    gint i;
 
-  g_log_set_handler (G_LOG_DOMAIN,
-                     G_LOG_LEVEL_MESSAGE,
-                     gimp_message_func,
-                     NULL);
-  g_log_set_handler (NULL,
-                     G_LOG_LEVEL_MESSAGE,
-                     gimp_message_func,
-                     NULL);
+    for (i = 0; i < G_N_ELEMENTS (log_domains); i++)
+      g_log_set_handler (log_domains[i],
+                         G_LOG_LEVEL_MESSAGE,
+                         gimp_message_func,
+                         NULL);
+
+    g_log_set_handler (NULL,
+                       G_LOG_LEVEL_MESSAGE,
+                       gimp_message_func,
+                       NULL);
+  }
 
   if (gimp_debug_flags & GIMP_DEBUG_FATAL_WARNINGS)
     {
@@ -509,9 +614,20 @@ gimp_main (const GimpPlugInInfo *info,
       fatal_mask = g_log_set_always_fatal (G_LOG_FATAL_MASK);
       fatal_mask |= G_LOG_LEVEL_WARNING | G_LOG_LEVEL_CRITICAL;
       g_log_set_always_fatal (fatal_mask);
+
+      g_log_set_handler (NULL,
+                         G_LOG_LEVEL_WARNING | G_LOG_LEVEL_CRITICAL |
+                         G_LOG_LEVEL_ERROR | G_LOG_FLAG_FATAL,
+                         gimp_fatal_func, NULL);
+    }
+  else
+    {
+      g_log_set_handler (NULL,
+                         G_LOG_LEVEL_ERROR | G_LOG_FLAG_FATAL,
+                         gimp_fatal_func, NULL);
     }
 
-  if (strcmp (argv[4], "-query") == 0)
+  if (strcmp (argv[ARG_MODE], "-query") == 0)
     {
       if (PLUG_IN_INFO.init_proc)
         gp_has_init_write (_writechannel, NULL);
@@ -527,7 +643,7 @@ gimp_main (const GimpPlugInInfo *info,
       return EXIT_SUCCESS;
     }
 
-  if (strcmp (argv[4], "-init") == 0)
+  if (strcmp (argv[ARG_MODE], "-init") == 0)
     {
       if (gimp_debug_flags & GIMP_DEBUG_INIT)
         gimp_debug_stop ();
@@ -568,6 +684,11 @@ gimp_quit (void)
 {
   gimp_close ();
 
+#if defined G_OS_WIN32 && defined HAVE_EXCHNDL
+  if (plug_in_backtrace_path)
+    g_free (plug_in_backtrace_path);
+#endif
+
   exit (EXIT_SUCCESS);
 }
 
@@ -610,6 +731,9 @@ gimp_quit (void)
  * "&lt;Domain&gt;/Path/To/My/Menu"
  * (e.g. "&lt;Image&gt;/Filters/Render/Useless").
  *
+ * Note that registering a full (pre-2.2-style) menu path is
+ * deprecated and will cause a failure in GIMP 3.0 and newer.
+ *
  * It is possible to register a procedure only for keyboard-shortcut
  * activation by passing a @menu_label to gimp_install_procedure() but
  * not registering any menu path with gimp_plugin_menu_register(). In
@@ -869,7 +993,7 @@ gimp_run_procedure (const gchar *name,
           (void) va_arg (args, gchar **);
           break;
         case GIMP_PDB_COLOR:
-	case GIMP_PDB_COLORARRAY:
+        case GIMP_PDB_COLORARRAY:
           (void) va_arg (args, GimpRGB *);
           break;
         case GIMP_PDB_PARASITE:
@@ -1054,10 +1178,12 @@ gimp_run_procedure2 (const gchar     *name,
   proc_run.nparams = n_params;
   proc_run.params  = (GPParam *) params;
 
+  gp_lock ();
   if (! gp_proc_run_write (_writechannel, &proc_run, NULL))
     gimp_quit ();
 
   gimp_read_expect_msg (&msg, GP_PROC_RETURN);
+  gp_unlock ();
 
   proc_return = msg.data;
 
@@ -1117,7 +1243,7 @@ gimp_destroy_paramdefs (GimpParamDef *paramdefs,
  *
  * If a procedure call fails, then it might pass an error message with
  * the return values. Plug-ins that are using the libgimp C wrappers
- * don't access the procedure return values directly. Thus ligimp
+ * don't access the procedure return values directly. Thus libgimp
  * stores the error message and makes it available with this
  * function. The next procedure call unsets the error message again.
  *
@@ -1126,7 +1252,7 @@ gimp_destroy_paramdefs (GimpParamDef *paramdefs,
  *
  * Return value: the error message
  *
- * Since: GIMP 2.6
+ * Since: 2.6
  **/
 const gchar *
 gimp_get_pdb_error (void)
@@ -1157,6 +1283,21 @@ gimp_get_pdb_error (void)
     }
 }
 
+/**
+ * gimp_get_pdb_status:
+ *
+ * Retrieves the status from the last procedure call.
+ *
+ * Return value: the #GimpPDBStatusType.
+ *
+ * Since: 2.10
+ **/
+GimpPDBStatusType
+gimp_get_pdb_status (void)
+{
+  return pdb_error_status;
+}
+
 /**
  * gimp_tile_width:
  *
@@ -1225,18 +1366,22 @@ gimp_shm_addr (void)
  * Returns the global gamma value GIMP and all its plug-ins should
  * use.
  *
- * This is a constant value given at plug-in configuration time.
+ * This is a constant value.
  *
  * NOTE: This function will always return 2.2, the gamma value for
- * sRGB. There's currently no way to change this and all operations
- * should assume that pixel data is in the sRGB colorspace.
+ * sRGB. If you need the actual gamma value of a drawable, look at its
+ * format.
+ *
+ * See also: gimp_drawable_get_format().
+ *
+ * @Deprecated: 2.8.4
  *
  * Return value: the gamma value
  **/
 gdouble
 gimp_gamma (void)
 {
-  return _gamma_val;
+  return 2.2;
 }
 
 /**
@@ -1254,7 +1399,7 @@ gimp_gamma (void)
 gboolean
 gimp_install_cmap (void)
 {
-  return _install_cmap;
+  return FALSE;
 }
 
 /**
@@ -1302,7 +1447,7 @@ gimp_show_tool_tips (void)
  *
  * Return value: the show_help_button boolean
  *
- * Since: GIMP 2.2
+ * Since: 2.2
  **/
 gboolean
 gimp_show_help_button (void)
@@ -1310,6 +1455,73 @@ gimp_show_help_button (void)
   return _show_help_button;
 }
 
+/**
+ * gimp_export_color_profile:
+ *
+ * Returns whether file plug-ins should default to exporting the
+ * image's color profile.
+ *
+ * Return value: TRUE if preferences are set to export the color profile.
+ *
+ * Since: 2.10.4
+ **/
+gboolean
+gimp_export_color_profile (void)
+{
+  return _export_profile;
+}
+
+/**
+ * gimp_export_exif:
+ *
+ * Returns whether file plug-ins should default to exporting Exif
+ * metadata, according preferences (original settings is #FALSE since
+ * metadata can contain sensitive information).
+ *
+ * Return value: TRUE if preferences are set to export Exif.
+ *
+ * Since: 2.10
+ **/
+gboolean
+gimp_export_exif (void)
+{
+  return _export_exif;
+}
+
+/**
+ * gimp_export_xmp:
+ *
+ * Returns whether file plug-ins should default to exporting XMP
+ * metadata, according preferences (original settings is #FALSE since
+ * metadata can contain sensitive information).
+ *
+ * Return value: TRUE if preferences are set to export XMP.
+ *
+ * Since: 2.10
+ **/
+gboolean
+gimp_export_xmp (void)
+{
+  return _export_xmp;
+}
+
+/**
+ * gimp_export_iptc:
+ *
+ * Returns whether file plug-ins should default to exporting IPTC
+ * metadata, according preferences (original settings is #FALSE since
+ * metadata can contain sensitive information).
+ *
+ * Return value: TRUE if preferences are set to export IPTC.
+ *
+ * Since: 2.10
+ **/
+gboolean
+gimp_export_iptc (void)
+{
+  return _export_iptc;
+}
+
 /**
  * gimp_check_size:
  *
@@ -1319,7 +1531,7 @@ gimp_show_help_button (void)
  *
  * Return value: the check_size value
  *
- * Since: GIMP 2.2
+ * Since: 2.2
  **/
 GimpCheckSize
 gimp_check_size (void)
@@ -1336,7 +1548,7 @@ gimp_check_size (void)
  *
  * Return value: the check_type value
  *
- * Since: GIMP 2.2
+ * Since: 2.2
  **/
 GimpCheckType
 gimp_check_type (void)
@@ -1418,7 +1630,7 @@ gimp_monitor_number (void)
  *
  * Return value: timestamp for plug-in window
  *
- * Since: GIMP 2.6
+ * Since: 2.6
  **/
 guint32
 gimp_user_time (void)
@@ -1426,6 +1638,23 @@ gimp_user_time (void)
   return _timestamp;
 }
 
+/**
+ * gimp_get_icon_theme_dir:
+ *
+ * Returns the directory of the current icon theme.
+ *
+ * This is a constant value given at plug-in configuration time.
+ *
+ * Return value: the icon theme directory
+ *
+ * Since: 2.10.4
+ **/
+const gchar *
+gimp_icon_theme_dir (void)
+{
+  return _icon_theme_dir;
+}
+
 /**
  * gimp_get_progname:
  *
@@ -1528,26 +1757,26 @@ gimp_extension_process (guint timeout)
       struct timeval *tvp;
 
       if (timeout)
-	{
-	  tv.tv_sec  = timeout / 1000;
-	  tv.tv_usec = (timeout % 1000) * 1000;
-	  tvp = &tv;
-	}
+        {
+          tv.tv_sec  = timeout / 1000;
+          tv.tv_usec = (timeout % 1000) * 1000;
+          tvp = &tv;
+        }
       else
-	tvp = NULL;
+        tvp = NULL;
 
       FD_ZERO (&readfds);
       FD_SET (g_io_channel_unix_get_fd (_readchannel), &readfds);
 
       if ((select_val = select (FD_SETSIZE, &readfds, NULL, NULL, tvp)) > 0)
-	{
-	  gimp_single_message ();
-	}
+        {
+          gimp_single_message ();
+        }
       else if (select_val == -1 && errno != EINTR)
-	{
-	  perror ("gimp_extension_process");
-	  gimp_quit ();
-	}
+        {
+          perror ("gimp_extension_process");
+          gimp_quit ();
+        }
     }
   while (select_val == -1 && errno == EINTR);
 #else
@@ -1696,10 +1925,51 @@ static void
 gimp_debug_stop (void)
 {
 #ifndef G_OS_WIN32
+
   g_log (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "Waiting for debugger...");
   raise (SIGSTOP);
+
 #else
-  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "Debugging not implemented on Win32");
+
+  HANDLE        hThreadSnap = NULL;
+  THREADENTRY32 te32        = { 0 };
+  pid_t         opid        = getpid ();
+
+  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG,
+         "Debugging (restart externally): %ld",
+         (long int) opid);
+
+  hThreadSnap = CreateToolhelp32Snapshot (TH32CS_SNAPTHREAD, 0);
+  if (hThreadSnap == INVALID_HANDLE_VALUE)
+    {
+      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG,
+             "error getting threadsnap - debugging impossible");
+      return;
+    }
+
+  te32.dwSize = sizeof (THREADENTRY32);
+
+  if (Thread32First (hThreadSnap, &te32))
+    {
+      do
+        {
+          if (te32.th32OwnerProcessID == opid)
+            {
+              HANDLE hThread = OpenThread (THREAD_SUSPEND_RESUME, FALSE,
+                                           te32.th32ThreadID);
+              SuspendThread (hThread);
+              CloseHandle (hThread);
+            }
+        }
+      while (Thread32Next (hThreadSnap, &te32));
+    }
+  else
+    {
+      g_log (G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "error getting threads");
+    }
+
+  CloseHandle (hThreadSnap);
+
 #endif
 }
 
@@ -1712,7 +1982,108 @@ gimp_message_func (const gchar    *log_domain,
   gimp_message (message);
 }
 
+static void
+gimp_fatal_func (const gchar    *log_domain,
+                 GLogLevelFlags  flags,
+                 const gchar    *message,
+                 gpointer        data)
+{
+  const gchar *level;
+
+  switch (flags & G_LOG_LEVEL_MASK)
+    {
+    case G_LOG_LEVEL_WARNING:
+      level = "WARNING";
+      break;
+    case G_LOG_LEVEL_CRITICAL:
+      level = "CRITICAL";
+      break;
+    case G_LOG_LEVEL_ERROR:
+      level = "ERROR";
+      break;
+    default:
+      level = "FATAL";
+      break;
+    }
+
+  g_printerr ("%s: %s: %s\n",
+              progname, level, message);
+
 #ifndef G_OS_WIN32
+  switch (stack_trace_mode)
+    {
+    case GIMP_STACK_TRACE_NEVER:
+      break;
+
+    case GIMP_STACK_TRACE_QUERY:
+        {
+          sigset_t sigset;
+
+          sigemptyset (&sigset);
+          sigprocmask (SIG_SETMASK, &sigset, NULL);
+          gimp_stack_trace_query (progname);
+        }
+      break;
+
+    case GIMP_STACK_TRACE_ALWAYS:
+        {
+          sigset_t sigset;
+
+          sigemptyset (&sigset);
+          sigprocmask (SIG_SETMASK, &sigset, NULL);
+          gimp_stack_trace_print (progname, stdout, NULL);
+        }
+      break;
+    }
+#endif
+
+  /* Do not end with gimp_quit().
+   * We want the plug-in to continue its normal crash course, otherwise
+   * we won't get the "Plug-in crashed" error in GIMP.
+   */
+  exit (EXIT_FAILURE);
+}
+
+#ifdef G_OS_WIN32
+
+#ifdef HAVE_EXCHNDL
+static LONG WINAPI
+gimp_plugin_sigfatal_handler (PEXCEPTION_POINTERS pExceptionInfo)
+{
+  g_printerr ("%s: fatal error\n", progname);
+
+  SetUnhandledExceptionFilter (_prevExceptionFilter);
+
+  /* For simplicity, do not make a difference between QUERY and ALWAYS
+   * on Windows (at least not for now).
+   */
+  if (stack_trace_mode != GIMP_STACK_TRACE_NEVER &&
+      g_file_test (plug_in_backtrace_path, G_FILE_TEST_IS_REGULAR))
+    {
+      FILE   *stream;
+      guchar  buffer[256];
+      size_t  read_len;
+
+      stream = fopen (plug_in_backtrace_path, "r");
+      do
+        {
+          /* Just read and output directly the file content. */
+          read_len = fread (buffer, 1, sizeof (buffer) - 1, stream);
+          buffer[read_len] = '\0';
+          g_printerr ("%s", buffer);
+        }
+      while (read_len);
+      fclose (stream);
+    }
+
+  if (_prevExceptionFilter && _prevExceptionFilter != gimp_plugin_sigfatal_handler)
+    return _prevExceptionFilter (pExceptionInfo);
+  else
+    return EXCEPTION_CONTINUE_SEARCH;
+}
+#endif
+
+#else
 static void
 gimp_plugin_sigfatal_handler (gint sig_num)
 {
@@ -1721,11 +2092,11 @@ gimp_plugin_sigfatal_handler (gint sig_num)
     case SIGHUP:
     case SIGINT:
     case SIGQUIT:
-    case SIGABRT:
     case SIGTERM:
       g_printerr ("%s terminated: %s\n", progname, g_strsignal (sig_num));
       break;
 
+    case SIGABRT:
     case SIGBUS:
     case SIGSEGV:
     case SIGFPE:
@@ -1743,7 +2114,7 @@ gimp_plugin_sigfatal_handler (gint sig_num)
 
             sigemptyset (&sigset);
             sigprocmask (SIG_SETMASK, &sigset, NULL);
-            g_on_error_query (progname);
+            gimp_stack_trace_query (progname);
           }
           break;
 
@@ -1753,14 +2124,18 @@ gimp_plugin_sigfatal_handler (gint sig_num)
 
             sigemptyset (&sigset);
             sigprocmask (SIG_SETMASK, &sigset, NULL);
-            g_on_error_stack_trace (progname);
+            gimp_stack_trace_print (progname, stdout, NULL);
           }
           break;
         }
       break;
     }
 
-  gimp_quit ();
+  /* Do not end with gimp_quit().
+   * We want the plug-in to continue its normal crash course, otherwise
+   * we won't get the "Plug-in crashed" error in GIMP.
+   */
+  exit (EXIT_FAILURE);
 }
 #endif
 
@@ -1922,6 +2297,9 @@ gimp_loop (void)
 static void
 gimp_config (GPConfig *config)
 {
+  GFile *file;
+  gchar *path;
+
   if (config->version < GIMP_PROTOCOL_VERSION)
     {
       g_message ("Could not execute plug-in \"%s\"\n(%s)\n"
@@ -1946,21 +2324,39 @@ gimp_config (GPConfig *config)
   _shm_ID           = config->shm_ID;
   _check_size       = config->check_size;
   _check_type       = config->check_type;
-  _install_cmap     = config->install_cmap     ? TRUE : FALSE;
   _show_tool_tips   = config->show_tooltips    ? TRUE : FALSE;
   _show_help_button = config->show_help_button ? TRUE : FALSE;
+  _export_profile   = config->export_profile   ? TRUE : FALSE;
+  _export_exif      = config->export_exif      ? TRUE : FALSE;
+  _export_xmp       = config->export_xmp       ? TRUE : FALSE;
+  _export_iptc      = config->export_iptc      ? TRUE : FALSE;
   _min_colors       = config->min_colors;
   _gdisp_ID         = config->gdisp_ID;
   _wm_class         = g_strdup (config->wm_class);
   _display_name     = g_strdup (config->display_name);
   _monitor_number   = config->monitor_number;
   _timestamp        = config->timestamp;
+  _icon_theme_dir   = g_strdup (config->icon_theme_dir);
 
   if (config->app_name)
     g_set_application_name (config->app_name);
 
   gimp_cpu_accel_set_use (config->use_cpu_accel);
 
+  file = gimp_file_new_for_config_path (config->swap_path, NULL);
+  path = g_file_get_path (file);
+
+  g_object_set (gegl_config (),
+                "tile-cache-size",     config->tile_cache_size,
+                "swap",                path,
+                "threads",             (gint) config->num_processors,
+                "use-opencl",          config->use_opencl,
+                "application-license", "GPL3",
+                NULL);
+
+  g_free (path);
+  g_object_unref (file);
+
   if (_shm_ID != -1)
     {
 #if defined(USE_SYSV_SHM)
@@ -1997,13 +2393,13 @@ gimp_config (GPConfig *config)
           /* Verify that we mapped our view */
           if (!_shm_addr)
             {
-              g_error ("MapViewOfFile error: %d... " ERRMSG_SHM_FAILED,
+              g_error ("MapViewOfFile error: %lu... " ERRMSG_SHM_FAILED,
                        GetLastError ());
             }
         }
       else
         {
-          g_error ("OpenFileMapping error: %d... " ERRMSG_SHM_FAILED,
+          g_error ("OpenFileMapping error: %lu... " ERRMSG_SHM_FAILED,
                    GetLastError ());
         }
 
