@@ -16,7 +16,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library.  If not, see
- * <http://www.gnu.org/licenses/>.
+ * <https://www.gnu.org/licenses/>.
  */
 
 #include "config.h"
@@ -33,7 +33,7 @@
 #include <AppKit/AppKit.h>
 #endif
 
-#include <glib-object.h>
+#include <gio/gio.h>
 #include <glib/gstdio.h>
 
 #undef GIMP_DISABLE_DEPRECATED
@@ -64,6 +64,14 @@
 #define gid_t gint
 #define geteuid() 0
 #define getegid() 0
+
+#include <shlobj.h>
+
+/* Constant available since Shell32.dll 4.72 */
+#ifndef CSIDL_APPDATA
+#define CSIDL_APPDATA 0x001a
+#endif
+
 #endif
 
 
@@ -77,8 +85,12 @@
  **/
 
 
-static gchar * gimp_env_get_dir (const gchar *gimp_env_name,
-                                 const gchar *env_dir);
+static gchar * gimp_env_get_dir   (const gchar *gimp_env_name,
+                                   const gchar *compile_time_dir,
+                                   const gchar *relative_subdir);
+#ifdef G_OS_WIN32
+static gchar * get_special_folder (gint         csidl);
+#endif
 
 
 const guint gimp_major_version = GIMP_MAJOR_VERSION;
@@ -94,7 +106,7 @@ const guint gimp_micro_version = GIMP_MICRO_VERSION;
  * you automatically (by means of the MAIN() macro that every plug-in
  * runs). Calling it again will cause a fatal error.
  *
- * Since: GIMP 2.4
+ * Since: 2.4
  */
 void
 gimp_env_init (gboolean plug_in)
@@ -117,7 +129,9 @@ gimp_env_init (gboolean plug_in)
       /* Set $LD_LIBRARY_PATH to ensure that plugins can be loaded. */
 
       const gchar *ldpath = g_getenv ("LD_LIBRARY_PATH");
-      gchar       *libdir = _gimp_reloc_find_lib_dir (NULL);
+      gchar       *libdir = g_build_filename (gimp_installation_directory (),
+                                              "lib",
+                                              NULL);
 
       if (ldpath && *ldpath)
         {
@@ -162,15 +176,29 @@ gimp_env_init (gboolean plug_in)
  * to be a subdirectory of gimp_data_directory().
  *
  * The usual case is that no GIMP2_DIRECTORY environment variable
- * exists, and then we use the GIMPDIR subdirectory of the home
- * directory. If no home directory exists, we use a per-user
- * subdirectory of gimp_data_directory().  In any case, we always
- * return some non-empty string, whether it corresponds to an existing
- * directory or not.
+ * exists, and then we use the GIMPDIR subdirectory of the local
+ * configuration directory:
+ *
+ * - UNIX: $XDG_CONFIG_HOME (defaults to $HOME/.config/)
+ *
+ * - Windows: CSIDL_APPDATA
+ *
+ * - OSX (UNIX exception): the Application Support Directory.
+ *
+ * If neither the configuration nor home directory exist,
+ * g_get_user_config_dir() will return {tmp}/{user_name}/.config/ where
+ * the temporary directory {tmp} and the {user_name} are determined
+ * according to platform rules.
+ *
+ * In any case, we always return some non-empty string, whether it
+ * corresponds to an existing directory or not.
+ *
+ * In config files such as gimprc, the string ${gimp_dir} expands to
+ * this directory.
  *
  * The returned string is owned by GIMP and must not be modified or
  * freed. The returned string is in the encoding used for filenames by
- * GLib, which isn't necessarily UTF-8. (On Windows it always is
+ * GLib, which isn't necessarily UTF-8 (on Windows it is always
  * UTF-8.)
  *
  * Returns: The user-specific GIMP settings directory.
@@ -182,7 +210,6 @@ gimp_directory (void)
   static gchar *last_env_gimp_dir = NULL;
 
   const gchar  *env_gimp_dir;
-  const gchar  *home_dir;
 
   env_gimp_dir = g_getenv ("GIMP2_DIRECTORY");
 
@@ -221,8 +248,6 @@ gimp_directory (void)
   g_free (last_env_gimp_dir);
   last_env_gimp_dir = g_strdup (env_gimp_dir);
 
-  home_dir = g_get_home_dir ();
-
   if (env_gimp_dir)
     {
       if (g_path_is_absolute (env_gimp_dir))
@@ -231,19 +256,18 @@ gimp_directory (void)
         }
       else
         {
+          const gchar *home_dir = g_get_home_dir ();
+
           if (home_dir)
-            {
-              gimp_dir = g_build_filename (home_dir,
-                                           env_gimp_dir,
-                                           NULL);
-            }
+            gimp_dir = g_build_filename (home_dir, env_gimp_dir, NULL);
           else
-            {
-              gimp_dir = g_build_filename (gimp_data_directory (),
-                                           env_gimp_dir, NULL);
-            }
+            gimp_dir = g_build_filename (gimp_data_directory (), env_gimp_dir, NULL);
         }
     }
+  else if (g_path_is_absolute (GIMPDIR))
+    {
+      gimp_dir = g_strdup (GIMPDIR);
+    }
   else
     {
 #ifdef PLATFORM_OSX
@@ -259,49 +283,25 @@ gimp_directory (void)
       library_dir = [path objectAtIndex:0];
 
       gimp_dir = g_build_filename ([library_dir UTF8String],
-                                   "GIMP", GIMP_USER_VERSION,
-                                   NULL);
+                                   GIMPDIR, GIMP_USER_VERSION, NULL);
 
       [pool drain];
 
-#else /* ! PLATFORM_OSX */
+#elif defined G_OS_WIN32
 
-      if (home_dir)
-        {
-          gimp_dir = g_build_filename (home_dir, GIMPDIR, NULL);
-        }
-      else
-        {
-          gchar *user_name = g_strdup (g_get_user_name ());
-          gchar *subdir_name;
+      gchar *conf_dir = get_special_folder (CSIDL_APPDATA);
 
-#ifdef G_OS_WIN32
-          gchar *p = user_name;
+      gimp_dir = g_build_filename (conf_dir,
+                                   GIMPDIR, GIMP_USER_VERSION, NULL);
+      g_free(conf_dir);
 
-          while (*p)
-            {
-              /* Replace funny characters in the user name with an
-               * underscore. The code below also replaces some
-               * characters that in fact are legal in file names, but
-               * who cares, as long as the definitely illegal ones are
-               * caught.
-               */
-              if (!g_ascii_isalnum (*p) && !strchr ("-.,@=", *p))
-                *p = '_';
-              p++;
-            }
-#endif
+#else /* UNIX */
 
-#ifndef G_OS_WIN32
-          g_message ("warning: no home directory.");
-#endif
-          subdir_name = g_strconcat (GIMPDIR ".", user_name, NULL);
-          gimp_dir = g_build_filename (gimp_data_directory (),
-                                       subdir_name,
-                                       NULL);
-          g_free (user_name);
-          g_free (subdir_name);
-        }
+      /* g_get_user_config_dir () always returns a path as a non-null
+       * and non-empty string
+       */
+      gimp_dir = g_build_filename (g_get_user_config_dir (),
+                                   GIMPDIR, GIMP_USER_VERSION, NULL);
 
 #endif /* PLATFORM_OSX */
     }
@@ -311,19 +311,41 @@ gimp_directory (void)
 
 #ifdef G_OS_WIN32
 
+/* Taken from glib 2.35 code. */
+static gchar *
+get_special_folder (int csidl)
+{
+  wchar_t      path[MAX_PATH+1];
+  HRESULT      hr;
+  LPITEMIDLIST pidl = NULL;
+  BOOL         b;
+  gchar       *retval = NULL;
+
+  hr = SHGetSpecialFolderLocation (NULL, csidl, &pidl);
+  if (hr == S_OK)
+    {
+      b = SHGetPathFromIDListW (pidl, path);
+      if (b)
+        retval = g_utf16_to_utf8 (path, -1, NULL, NULL, NULL);
+      CoTaskMemFree (pidl);
+    }
+
+  return retval;
+}
+
 static HMODULE libgimpbase_dll = NULL;
 
 /* Minimal DllMain that just stores the handle to this DLL */
 
-BOOL WINAPI			/* Avoid silly "no previous prototype" gcc warning */
+BOOL WINAPI /* Avoid silly "no previous prototype" gcc warning */
 DllMain (HINSTANCE hinstDLL,
-	 DWORD     fdwReason,
-	 LPVOID    lpvReserved);
+         DWORD     fdwReason,
+         LPVOID    lpvReserved);
 
 BOOL WINAPI
 DllMain (HINSTANCE hinstDLL,
-	 DWORD     fdwReason,
-	 LPVOID    lpvReserved)
+         DWORD     fdwReason,
+         LPVOID    lpvReserved)
 {
   switch (fdwReason)
     {
@@ -346,12 +368,15 @@ DllMain (HINSTANCE hinstDLL,
  * filename is used. On OSX we ask [NSBundle mainBundle] for the
  * resource path to check if GIMP is part of a relocatable bundle.
  *
+ * In config files such as gimprc, the string ${gimp_installation_dir}
+ * expands to this directory.
+ *
  * The returned string is owned by GIMP and must not be modified or
  * freed. The returned string is in the encoding used for filenames by
  * GLib, which isn't necessarily UTF-8. (On Windows it always is
  * UTF-8.)
  *
- * Since: GIMP 2.8
+ * Since: 2.8
  *
  * Returns: The toplevel installation directory of GIMP.
  **/
@@ -365,39 +390,9 @@ gimp_installation_directory (void)
 
 #ifdef G_OS_WIN32
 
-  {
-    /* Figure it out from the location of this DLL */
-    gchar *filename;
-    gchar *sep1, *sep2;
-
-    wchar_t w_filename[MAX_PATH];
-
-    if (GetModuleFileNameW (libgimpbase_dll, w_filename, G_N_ELEMENTS (w_filename)) == 0)
-      g_error ("GetModuleFilenameW failed");
-
-    filename = g_utf16_to_utf8 (w_filename, -1, NULL, NULL, NULL);
-    if (filename == NULL)
-      g_error ("Converting module filename to UTF-8 failed");
-
-    /* If the DLL file name is of the format
-     * <foobar>\bin\*.dll, use <foobar>.
-     * Otherwise, use the directory where the DLL is.
-     */
-
-    sep1 = strrchr (filename, '\\');
-    *sep1 = '\0';
-
-    sep2 = strrchr (filename, '\\');
-    if (sep2 != NULL)
-      {
-        if (g_ascii_strcasecmp (sep2 + 1, "bin") == 0)
-          {
-            *sep2 = '\0';
-          }
-      }
-
-    toplevel = filename;
-  }
+  toplevel = g_win32_get_package_installation_directory_of_module (libgimpbase_dll);
+  if (! toplevel)
+    g_error ("g_win32_get_package_installation_directory_of_module() failed");
 
 #elif PLATFORM_OSX
 
@@ -465,11 +460,17 @@ gimp_installation_directory (void)
 /**
  * gimp_data_directory:
  *
- * Returns the top directory for GIMP data. If the environment
+ * Returns the default top directory for GIMP data. If the environment
  * variable GIMP2_DATADIR exists, that is used.  It should be an
  * absolute pathname.  Otherwise, on Unix the compile-time defined
- * directory is used. On Windows, the installation directory as deduced
- * from the executable's full filename is used.
+ * directory is used. On Windows, the installation directory as
+ * deduced from the executable's full filename is used.
+ *
+ * Note that the actual directories used for GIMP data files can be
+ * overridden by the user in the preferences dialog.
+ *
+ * In config files such as gimprc, the string ${gimp_data_dir} expands
+ * to this directory.
  *
  * The returned string is owned by GIMP and must not be modified or
  * freed. The returned string is in the encoding used for filenames by
@@ -485,9 +486,12 @@ gimp_data_directory (void)
 
   if (! gimp_data_dir)
     {
-      gchar *tmp = _gimp_reloc_find_data_dir (DATADIR);
+      gchar *tmp = g_build_filename ("share",
+                                     GIMP_PACKAGE,
+                                     GIMP_DATA_VERSION,
+                                     NULL);
 
-      gimp_data_dir = gimp_env_get_dir ("GIMP2_DATADIR", tmp);
+      gimp_data_dir = gimp_env_get_dir ("GIMP2_DATADIR", GIMPDATADIR, tmp);
       g_free (tmp);
     }
 
@@ -520,11 +524,20 @@ gimp_locale_directory (void)
 
   if (! gimp_locale_dir)
     {
-      gchar *tmp = _gimp_reloc_find_locale_dir (LOCALEDIR);
+      gchar *tmp = g_build_filename ("share",
+                                     "locale",
+                                     NULL);
 
-      gimp_locale_dir = gimp_env_get_dir ("GIMP2_LOCALEDIR", tmp);
+      gimp_locale_dir = gimp_env_get_dir ("GIMP2_LOCALEDIR", LOCALEDIR, tmp);
       g_free (tmp);
+
 #ifdef G_OS_WIN32
+      /* FIXME: g_win32_locale_filename_from_utf8() can actually return
+       * NULL (we had actual cases of this). Not sure exactly what
+       * gimp_locale_directory() should do when this happens. Anyway
+       * that's really broken, and something should be done some day
+       * about this!
+       */
       tmp = g_win32_locale_filename_from_utf8 (gimp_locale_dir);
       g_free (gimp_locale_dir);
       gimp_locale_dir = tmp;
@@ -543,6 +556,9 @@ gimp_locale_directory (void)
  * directory is used. On Windows, the installation directory as deduced
  * from the executable's full filename is used.
  *
+ * In config files such as gimprc, the string ${gimp_sysconf_dir}
+ * expands to this directory.
+ *
  * The returned string is owned by GIMP and must not be modified or
  * freed. The returned string is in the encoding used for filenames by
  * GLib, which isn't necessarily UTF-8. (On Windows it always is
@@ -557,40 +573,32 @@ gimp_sysconf_directory (void)
 
   if (! gimp_sysconf_dir)
     {
-      gchar *tmp = _gimp_reloc_find_etc_dir (SYSCONFDIR);
+      gchar *tmp = g_build_filename ("etc",
+                                     GIMP_PACKAGE,
+                                     GIMP_SYSCONF_VERSION,
+                                     NULL);
 
-      gimp_sysconf_dir = gimp_env_get_dir ("GIMP2_SYSCONFDIR", tmp);
+      gimp_sysconf_dir = gimp_env_get_dir ("GIMP2_SYSCONFDIR", GIMPSYSCONFDIR, tmp);
       g_free (tmp);
     }
 
   return gimp_sysconf_dir;
 }
 
-/**
- * gimp_user_directory:
- * @type: the type of user directory to retrieve
- *
- * This procedure is deprecated! Use g_get_user_special_dir() instead.
- *
- * Returns: The path to the specified user directory, or %NULL if the
- *          logical ID was not found.
- *
- * Since: GIMP 2.4
- **/
-const gchar *
-gimp_user_directory (GimpUserDirectory type)
-{
-  return g_get_user_special_dir (type);
-}
-
 /**
  * gimp_plug_in_directory:
  *
- * Returns the top directory for GIMP plug_ins and modules. If the
- * environment variable GIMP2_PLUGINDIR exists, that is used.  It
+ * Returns the default top directory for GIMP plug-ins and modules. If
+ * the environment variable GIMP2_PLUGINDIR exists, that is used.  It
  * should be an absolute pathname. Otherwise, on Unix the compile-time
- * defined directory is used. On Windows, the installation directory as
- * deduced from the executable's full filename is used.
+ * defined directory is used. On Windows, the installation directory
+ * as deduced from the executable's full filename is used.
+ *
+ * Note that the actual directories used for GIMP plug-ins and modules
+ * can be overridden by the user in the preferences dialog.
+ *
+ * In config files such as gimprc, the string ${gimp_plug_in_dir}
+ * expands to this directory.
  *
  * The returned string is owned by GIMP and must not be modified or
  * freed. The returned string is in the encoding used for filenames by
@@ -606,25 +614,329 @@ gimp_plug_in_directory (void)
 
   if (! gimp_plug_in_dir)
     {
-      gchar *tmp = _gimp_reloc_find_plugin_dir (PLUGINDIR);
+      gchar *tmp = g_build_filename ("lib",
+                                     GIMP_PACKAGE,
+                                     GIMP_PLUGIN_VERSION,
+                                     NULL);
 
-      gimp_plug_in_dir = gimp_env_get_dir ("GIMP2_PLUGINDIR", tmp);
+      gimp_plug_in_dir = gimp_env_get_dir ("GIMP2_PLUGINDIR", PLUGINDIR, tmp);
       g_free (tmp);
     }
 
   return gimp_plug_in_dir;
 }
 
+/**
+ * gimp_cache_directory:
+ *
+ * Returns the default top directory for GIMP cached files. If the
+ * environment variable GIMP2_CACHEDIR exists, that is used.  It
+ * should be an absolute pathname.  Otherwise, a subdirectory of the
+ * directory returned by g_get_user_cache_dir() is used.
+ *
+ * Note that the actual directories used for GIMP caches files can
+ * be overridden by the user in the preferences dialog.
+ *
+ * In config files such as gimprc, the string ${gimp_cache_dir}
+ * expands to this directory.
+ *
+ * The returned string is owned by GIMP and must not be modified or
+ * freed. The returned string is in the encoding used for filenames by
+ * GLib, which isn't necessarily UTF-8. (On Windows it always is
+ * UTF-8.).
+ *
+ * Since: 2.10.10
+ *
+ * Returns: The default top directory for GIMP cached files.
+ **/
+const gchar *
+gimp_cache_directory (void)
+{
+  static gchar *gimp_cache_dir = NULL;
+
+  if (! gimp_cache_dir)
+    {
+      gchar *tmp = g_build_filename (g_get_user_cache_dir (),
+                                     GIMP_PACKAGE,
+                                     GIMP_USER_VERSION,
+                                     NULL);
+
+      gimp_cache_dir = gimp_env_get_dir ("GIMP2_CACHEDIR", NULL, tmp);
+      g_free (tmp);
+    }
+
+  return gimp_cache_dir;
+}
+
+/**
+ * gimp_temp_directory:
+ *
+ * Returns the default top directory for GIMP temporary files. If the
+ * environment variable GIMP2_TEMPDIR exists, that is used.  It
+ * should be an absolute pathname.  Otherwise, a subdirectory of the
+ * directory returned by g_get_tmp_dir() is used.
+ *
+ * In config files such as gimprc, the string ${gimp_temp_dir} expands
+ * to this directory.
+ *
+ * Note that the actual directories used for GIMP temporary files can
+ * be overridden by the user in the preferences dialog.
+ *
+ * The returned string is owned by GIMP and must not be modified or
+ * freed. The returned string is in the encoding used for filenames by
+ * GLib, which isn't necessarily UTF-8. (On Windows it always is
+ * UTF-8.).
+ *
+ * Since: 2.10.10
+ *
+ * Returns: The default top directory for GIMP temporary files.
+ **/
+const gchar *
+gimp_temp_directory (void)
+{
+  static gchar *gimp_temp_dir = NULL;
+
+  if (! gimp_temp_dir)
+    {
+      gchar *tmp = g_build_filename (g_get_tmp_dir (),
+                                     GIMP_PACKAGE,
+                                     GIMP_USER_VERSION,
+                                     NULL);
+
+      gimp_temp_dir = gimp_env_get_dir ("GIMP2_TEMPDIR", NULL, tmp);
+      g_free (tmp);
+    }
+
+  return gimp_temp_dir;
+}
+
+static GFile *
+gimp_child_file (const gchar *parent,
+                 const gchar *element,
+                 va_list      args)
+{
+  GFile *file = g_file_new_for_path (parent);
+
+  while (element)
+    {
+      GFile *child = g_file_get_child (file, element);
+
+      g_object_unref (file);
+      file = child;
+
+      element = va_arg (args, const gchar *);
+    }
+
+  return file;
+}
+
+/**
+ * gimp_directory_file:
+ * @first_element: the first element of a path to a file in the
+ *                 user's GIMP directory, or %NULL.
+ * @...: a %NULL terminated list of the remaining elements of the path
+ *       to the file.
+ *
+ * Returns a #GFile in the user's GIMP directory, or the data
+ * directory itself if @first_element is %NULL.
+ *
+ * See also: gimp_directory().
+ *
+ * Since: 2.10
+ *
+ * Returns: a new @GFile for the path, Free with g_object_unref().
+ **/
+GFile *
+gimp_directory_file (const gchar *first_element,
+                     ...)
+{
+  GFile   *file;
+  va_list  args;
+
+  va_start (args, first_element);
+  file = gimp_child_file (gimp_directory (), first_element, args);
+  va_end (args);
+
+  return file;
+}
+
+/**
+ * gimp_installation_directory_file:
+ * @first_element: the first element of a path to a file in the
+ *                 top installation directory, or %NULL.
+ * @...: a %NULL terminated list of the remaining elements of the path
+ *       to the file.
+ *
+ * Returns a #GFile in the installation directory, or the installation
+ * directory itself if @first_element is %NULL.
+ *
+ * See also: gimp_installation_directory().
+ *
+ * Since: 2.10.10
+ *
+ * Returns: a new @GFile for the path, Free with g_object_unref().
+ **/
+GFile *
+gimp_installation_directory_file (const gchar *first_element,
+                                  ...)
+{
+  GFile   *file;
+  va_list  args;
+
+  va_start (args, first_element);
+  file = gimp_child_file (gimp_installation_directory (), first_element, args);
+  va_end (args);
+
+  return file;
+}
+
+/**
+ * gimp_data_directory_file:
+ * @first_element: the first element of a path to a file in the
+ *                 data directory, or %NULL.
+ * @...: a %NULL terminated list of the remaining elements of the path
+ *       to the file.
+ *
+ * Returns a #GFile in the data directory, or the data directory
+ * itself if @first_element is %NULL.
+ *
+ * See also: gimp_data_directory().
+ *
+ * Since: 2.10
+ *
+ * Returns: a new @GFile for the path, Free with g_object_unref().
+ **/
+GFile *
+gimp_data_directory_file (const gchar *first_element,
+                          ...)
+{
+  GFile   *file;
+  va_list  args;
+
+  va_start (args, first_element);
+  file = gimp_child_file (gimp_data_directory (), first_element, args);
+  va_end (args);
+
+  return file;
+}
+
+/**
+ * gimp_locale_directory_file:
+ * @first_element: the first element of a path to a file in the
+ *                 locale directory, or %NULL.
+ * @...: a %NULL terminated list of the remaining elements of the path
+ *       to the file.
+ *
+ * Returns a #GFile in the locale directory, or the locale directory
+ * itself if @first_element is %NULL.
+ *
+ * See also: gimp_locale_directory().
+ *
+ * Since: 2.10
+ *
+ * Returns: a new @GFile for the path, Free with g_object_unref().
+ **/
+GFile *
+gimp_locale_directory_file (const gchar *first_element,
+                            ...)
+{
+  GFile   *file;
+  va_list  args;
+
+  va_start (args, first_element);
+  file = gimp_child_file (gimp_locale_directory (), first_element, args);
+  va_end (args);
+
+  return file;
+}
+
+/**
+ * gimp_sysconf_directory_file:
+ * @first_element: the first element of a path to a file in the
+ *                 sysconf directory, or %NULL.
+ * @...: a %NULL terminated list of the remaining elements of the path
+ *       to the file.
+ *
+ * Returns a #GFile in the sysconf directory, or the sysconf directory
+ * itself if @first_element is %NULL.
+ *
+ * See also: gimp_sysconf_directory().
+ *
+ * Since: 2.10
+ *
+ * Returns: a new @GFile for the path, Free with g_object_unref().
+ **/
+GFile *
+gimp_sysconf_directory_file (const gchar *first_element,
+                             ...)
+{
+  GFile   *file;
+  va_list  args;
+
+  va_start (args, first_element);
+  file = gimp_child_file (gimp_sysconf_directory (), first_element, args);
+  va_end (args);
+
+  return file;
+}
+
+/**
+ * gimp_plug_in_directory_file:
+ * @first_element: the first element of a path to a file in the
+ *                 plug-in directory, or %NULL.
+ * @...: a %NULL terminated list of the remaining elements of the path
+ *       to the file.
+ *
+ * Returns a #GFile in the plug-in directory, or the plug-in directory
+ * itself if @first_element is %NULL.
+ *
+ * See also: gimp_plug_in_directory().
+ *
+ * Since: 2.10
+ *
+ * Returns: a new @GFile for the path, Free with g_object_unref().
+ **/
+GFile *
+gimp_plug_in_directory_file (const gchar *first_element,
+                             ...)
+{
+  GFile   *file;
+  va_list  args;
+
+  va_start (args, first_element);
+  file = gimp_child_file (gimp_plug_in_directory (), first_element, args);
+  va_end (args);
+
+  return file;
+}
+
+/**
+ * gimp_user_directory:
+ * @type: the type of user directory to retrieve
+ *
+ * This procedure is deprecated! Use g_get_user_special_dir() instead.
+ *
+ * Returns: The path to the specified user directory, or %NULL if the
+ *          logical ID was not found.
+ *
+ * Since: 2.4
+ **/
+const gchar *
+gimp_user_directory (GimpUserDirectory type)
+{
+  return g_get_user_special_dir ((GUserDirectory) type);
+}
+
 /**
  * gimp_personal_rc_file:
  * @basename: The basename of a rc_file.
  *
  * Returns the name of a file in the user-specific GIMP settings directory.
  *
- * The returned string is allocated dynamically and *SHOULD* be freed
- * with g_free() after use. The returned string is in the encoding
- * used for filenames by GLib, which isn't necessarily
- * UTF-8. (On Windows it always is UTF-8.)
+ * The returned string is newly allocated and should be freed with
+ * g_free() after use. The returned string is in the encoding used for
+ * filenames by GLib, which isn't necessarily UTF-8. (On Windows it
+ * always is UTF-8.)
  *
  * Returns: The name of a file in the user-specific GIMP settings directory.
  **/
@@ -653,7 +965,7 @@ gimp_gtkrc (void)
 
   if (! gimp_gtkrc_filename)
     gimp_gtkrc_filename = g_build_filename (gimp_data_directory (),
-                                            "themes", "Default", "gtkrc",
+                                            "themes", "System", "gtkrc",
                                             NULL);
 
   return gimp_gtkrc_filename;
@@ -745,18 +1057,15 @@ gimp_path_parse (const gchar  *path,
                  gboolean      check,
                  GList       **check_failed)
 {
-  const gchar  *home;
-  gchar       **patharray;
-  GList        *list      = NULL;
-  GList        *fail_list = NULL;
-  gint          i;
-  gboolean      exists    = TRUE;
+  gchar    **patharray;
+  GList     *list      = NULL;
+  GList     *fail_list = NULL;
+  gint       i;
+  gboolean   exists    = TRUE;
 
   if (!path || !*path || max_paths < 1 || max_paths > 256)
     return NULL;
 
-  home = g_get_home_dir ();
-
   patharray = g_strsplit (path, G_SEARCHPATH_SEPARATOR_S, max_paths);
 
   for (i = 0; i < max_paths; i++)
@@ -769,7 +1078,7 @@ gimp_path_parse (const gchar  *path,
 #ifndef G_OS_WIN32
       if (*patharray[i] == '~')
         {
-          dir = g_string_new (home);
+          dir = g_string_new (g_get_home_dir ());
           g_string_append (dir, patharray[i] + 1);
         }
       else
@@ -783,9 +1092,24 @@ gimp_path_parse (const gchar  *path,
         exists = g_file_test (dir->str, G_FILE_TEST_IS_DIR);
 
       if (exists)
-        list = g_list_prepend (list, g_strdup (dir->str));
+        {
+          GList *dup;
+
+          /*  check for duplicate entries, see bug #784502  */
+          for (dup = list; dup; dup = g_list_next (dup))
+            {
+              if (! strcmp (dir->str, dup->data))
+                break;
+            }
+
+          /*  only add to the list if it's not a duplicate  */
+          if (! dup)
+            list = g_list_prepend (list, g_strdup (dir->str));
+        }
       else if (check_failed)
-        fail_list = g_list_prepend (fail_list, g_strdup (dir->str));
+        {
+          fail_list = g_list_prepend (fail_list, g_strdup (dir->str));
+        }
 
       g_string_free (dir, TRUE);
     }
@@ -904,7 +1228,8 @@ gimp_path_get_user_writable_dir (GList *path)
 
 static gchar *
 gimp_env_get_dir (const gchar *gimp_env_name,
-                  const gchar *env_dir)
+                  const gchar *compile_time_dir,
+                  const gchar *relative_subdir)
 {
   const gchar *env = g_getenv (gimp_env_name);
 
@@ -916,12 +1241,20 @@ gimp_env_get_dir (const gchar *gimp_env_name,
 
       return g_strdup (env);
     }
-  else
+  else if (compile_time_dir)
     {
-      gchar *retval = g_strdup (env_dir);
+      gchar *retval = g_strdup (compile_time_dir);
 
       gimp_path_runtime_fix (&retval);
 
       return retval;
     }
+  else if (! g_path_is_absolute (relative_subdir))
+    {
+      return g_build_filename (gimp_installation_directory (),
+                               relative_subdir,
+                               NULL);
+    }
+
+  return g_strdup (relative_subdir);
 }
